/**
 *
 * Copyright (c) 2001, Frank Bu
 *
 * project: Formula
 * version: $Revision: 1.3 $ $Name:  $
 *
 * Evaluator class and all Action classes.
 */

// disable "identifier was truncated to 255 characters in the debug information"
#pragma warning(disable : 4786)

#include "Evaluator.h"

using namespace std;

double NumberStack::popWithReturn()
{
	double value = top();
	pop();
	return value;
}


void Action::checkTopStackElement(NumberStack& numberStack)
{
	if (!_finite(numberStack.top()) || _isnan(numberStack.top())) throw MathError();
}


NumberAction::NumberAction(string value)
{
	m_value = atof(value.c_str());
}

void NumberAction::run(NumberStack& numberStack)
{
	numberStack.push(m_value);
	checkTopStackElement(numberStack);
}



void MulAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 * op2);
	checkTopStackElement(numberStack);
}


void DivAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 / op2);
	checkTopStackElement(numberStack);
}


void AddAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 + op2);
	checkTopStackElement(numberStack);
}

void LessAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 < op2);
	checkTopStackElement(numberStack);
}

void GreaterAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 > op2);
	checkTopStackElement(numberStack);
}

void LessEqualAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 <= op2);
	checkTopStackElement(numberStack);
}

void GreaterEqualAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 >= op2);
	checkTopStackElement(numberStack);
}

void EqualAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 == op2);
	checkTopStackElement(numberStack);
}

void NotEqualAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 != op2);
	checkTopStackElement(numberStack);
}

void AndAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 && op2);
	checkTopStackElement(numberStack);
}

void OrAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 || op2);
	checkTopStackElement(numberStack);
}

void NotAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 1) throw StackUnderflow();
	numberStack.push(!numberStack.popWithReturn());
	checkTopStackElement(numberStack);
}

void SubAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(op1 - op2);
	checkTopStackElement(numberStack);
}

void NegAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 1) throw StackUnderflow();
	numberStack.push(-numberStack.popWithReturn());
	checkTopStackElement(numberStack);
}

void PowerAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(pow(op1, op2));
	checkTopStackElement(numberStack);
}

Evaluator::~Evaluator()
{
	deleteActions();
}

void Evaluator::addAction(Action* action)
{
	m_actions.push_back(action);
}

double Evaluator::eval()
{
	if (m_actions.size() == 0) return 0;
	NumberStack numberStack;
	for (int i = 0; i < m_actions.size(); i++) m_actions[i]->run(numberStack);
	if (numberStack.size() != 1) throw StackUnderflow();
	return numberStack.popWithReturn();
}

void Evaluator::removeAllActions()
{
	deleteActions();
	m_actions.clear();
}

void Evaluator::setVariable(string name, double value)
{
	m_variables[name] = value;
}

void Evaluator::removeVariable(string name)
{
	map<string, double>::iterator i = m_variables.find(name);
	if (i != m_variables.end()) {
		m_variables.erase(i);
	}
}

void Evaluator::removeAllVariables()
{
	m_variables.clear();
}

double Evaluator::getVariable(string name)
{
	map<string, double>::iterator i = m_variables.find(name);
	if (i != m_variables.end()) {
		return i->second;
	} else {
		throw VariableNotFound(name);
	}
	return 0;
}

void Evaluator::setFunction(string name, double(*function)())
{
	m_noArgumentFunctions[name] = function;
}

void Evaluator::setFunction(string name, double(*function)(double))
{
	m_oneArgumentFunctions[name] = function;
}

void Evaluator::setFunction(string name, double(*function)(double, double))
{
	m_twoArgumentsFunctions[name] = function;
}

void Evaluator::removeFunction(string name)
{
	map<string, double(*)()>::iterator no = m_noArgumentFunctions.find(name);
	if (no != m_noArgumentFunctions.end()) {
		m_noArgumentFunctions.erase(no);
	}
	map<string, double(*)(double)>::iterator one = m_oneArgumentFunctions.find(name);
	if (one != m_oneArgumentFunctions.end()) {
		m_oneArgumentFunctions.erase(one);
	}
	map<string, double(*)(double, double)>::iterator two = m_twoArgumentsFunctions.find(name);
	if (two != m_twoArgumentsFunctions.end()) {
		m_twoArgumentsFunctions.erase(two);
	}
}

void Evaluator::removeAllFunctions()
{
	m_noArgumentFunctions.clear();
	m_oneArgumentFunctions.clear();
	m_twoArgumentsFunctions.clear();
}

double Evaluator::evalFunction(string name)
{
	double(*function)() = m_noArgumentFunctions[name];
	if (function) {
		return function();
	} else {
		throw FunctionNotFound(name);
	}
	return 0;
}

double Evaluator::evalFunction(string name, double argument)
{
	double(*function)(double) = m_oneArgumentFunctions[name];
	if (function) {
		return function(argument);
	} else {
		throw FunctionNotFound(name);
	}
	return 0;
}

double Evaluator::evalFunction(string name, double argument1, double argument2)
{
	double(*function)(double, double) = m_twoArgumentsFunctions[name];
	if (function) {
		return function(argument1, argument2);
	} else {
		throw FunctionNotFound(name);
	}
	return 0;
}


void Evaluator::deleteActions()
{
	for (int i = 0; i < m_actions.size(); i++) delete m_actions[i];
}


void VariableAction::run(NumberStack& numberStack)
{
	numberStack.push(m_evaluator->getVariable(m_name));
	checkTopStackElement(numberStack);
}


void NoArgumentFunctionAction::run(NumberStack& numberStack)
{
	numberStack.push(m_evaluator->evalFunction(m_name));
	checkTopStackElement(numberStack);
}

void OneArgumentFunctionAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 1) throw StackUnderflow();
	numberStack.push(m_evaluator->evalFunction(m_name, numberStack.popWithReturn()));
	checkTopStackElement(numberStack);
}

void TwoArgumentsFunctionAction::run(NumberStack& numberStack)
{
	if (numberStack.size() < 2) throw StackUnderflow();
	double op2 = numberStack.popWithReturn();
	double op1 = numberStack.popWithReturn();
	numberStack.push(m_evaluator->evalFunction(m_name, op1, op2));
	checkTopStackElement(numberStack);
}
