/**
 *
 * Copyright (c) 2001, Frank Bu
 *
 * project: Formula
 * version: $Revision: 1.3 $ $Name:  $
 *
 * Evaluator class and all Action classes.
 */

#ifndef EVALUATOR_H
#define EVALUATOR_H

#include <string>
#include <vector>
#include <stack>
#include <map>
#include <math.h>
#include <float.h>

#include "Exception.h"
#include "Exception.h"

using namespace std;

class NumberStack : public stack<double>
{
public:
	double popWithReturn();
};


class Action
{
public:
	virtual ~Action() {};
	virtual void run(NumberStack& numberStack) = 0;
protected:
	void checkTopStackElement(NumberStack& numberStack);
};


class NumberAction : public Action
{
public:
	NumberAction(double value) : m_value(value) {}
	NumberAction(string value);

	void run(NumberStack& numberStack);

private:
	double m_value;
};


class MulAction : public Action
{
public:
	void run(NumberStack& numberStack);
};


class DivAction : public Action
{
public:
	void run(NumberStack& numberStack);
};


class AddAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class LessAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class GreaterAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class LessEqualAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class GreaterEqualAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class EqualAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class NotEqualAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class AndAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class OrAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class NotAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class SubAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class NegAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class PowerAction : public Action
{
public:
	void run(NumberStack& numberStack);
};

class Evaluator
{
public:
	~Evaluator();
	void addAction(Action* action);
	double eval() throw(FunctionNotFound, VariableNotFound, StackUnderflow, MathError);
	void removeAllActions();
	void setVariable(string name, double value);
	void removeVariable(string name);
	void removeAllVariables();
	double getVariable(string name) throw(VariableNotFound);
	void setFunction(string name, double(*function)());
	void setFunction(string name, double(*function)(double));
	void setFunction(string name, double(*function)(double, double));
	void removeFunction(string name);
	void removeAllFunctions();
	double evalFunction(string name) throw(FunctionNotFound, StackUnderflow, MathError);
	double evalFunction(string name, double argument) throw(FunctionNotFound, StackUnderflow, MathError);
	double evalFunction(string name, double argument1, double argument2) throw(FunctionNotFound, StackUnderflow, MathError);

private:

	void deleteActions();

	vector<Action*> m_actions;
	map<string, double> m_variables;
	map<string, double(*)()> m_noArgumentFunctions;
	map<string, double(*)(double)> m_oneArgumentFunctions;
	map<string, double(*)(double, double)> m_twoArgumentsFunctions;
};


class VariableAction : public Action
{
public:
	VariableAction(Evaluator* evaluator, string name) : m_evaluator(evaluator), m_name(name) {}
	void run(NumberStack& numberStack) throw(VariableNotFound, MathError);

private:
	Evaluator* m_evaluator;
	string m_name;
};


class NoArgumentFunctionAction : public Action
{
public:
	NoArgumentFunctionAction(Evaluator* evaluator, string name) : m_evaluator(evaluator), m_name(name) {}
	void run(NumberStack& numberStack) throw(FunctionNotFound, MathError);

private:
	Evaluator* m_evaluator;
	string m_name;
};

class OneArgumentFunctionAction : public Action
{
public:
	OneArgumentFunctionAction(Evaluator* evaluator, string name) : m_evaluator(evaluator), m_name(name) {}
	void run(NumberStack& numberStack) throw(FunctionNotFound, StackUnderflow, MathError);

private:
	Evaluator* m_evaluator;
	string m_name;
};

class TwoArgumentsFunctionAction : public Action
{
public:
	TwoArgumentsFunctionAction(Evaluator* evaluator, string name) : m_evaluator(evaluator), m_name(name) {}
	void run(NumberStack& numberStack) throw(FunctionNotFound, StackUnderflow, MathError);

private:
	Evaluator* m_evaluator;
	string m_name;
};


#endif
